==> group_clone-1.0.x-dev.tar.gz: group_clone/src/GroupCloner.php <==
<?php

declare(strict_types=1);

namespace Drupal\group_clone;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\group_clone\Plugin\GroupContentCloner\GroupContentClonerInterface;
use Drupal\group\Entity\GroupContentInterface;
use Drupal\group\Entity\GroupInterface;

/**
 * {@inheritdoc}
 */
final class GroupCloner {

  use EntityCloneTrait;
  use StringTranslationTrait;

  public const STATE_PREFIX = 'group_clone_data.';

  /**
   * {@inheritdoc}
   */
  private GroupContentClonerManager $clonerManager;

  /**
   * {@inheritdoc}
   */
  private EntityTypeManagerInterface $entityTypeManager;

  /**
   * {@inheritdoc}
   */
  private StateInterface $state;

  /**
   * Cloning settings keyed by group content type.
   *
   * @var int[]
   */
  private array $settings = [];

  /**
   * Group content type - base plugin ID map.
   *
   * Data caching purpose.
   *
   * @var string[]
   */
  private array $pluginMap = [];

  /**
   * Cloner plugins.
   *
   * @var \Drupal\group_clone\Plugin\GroupContentCloner\GroupContentClonerInterface[]
   */
  private array $clonerPlugins = [];

  /**
   * Cloned entities array for referencing.
   *
   * Keyed by entity type ID as it can contain all types of entities
   * and not only group_content ones.
   *
   * @var mixed[]
   */
  private array $clonedEntityIDs = [];

  /**
   * Is cloning in progress?
   *
   * @var bool
   */
  private bool $cloning = FALSE;

  /**
   * Is reverting in progress?
   *
   * @var bool
   */
  private bool $reverting = FALSE;

  /**
   * Constructs a new GroupCloner.
   */
  public function __construct(
    GroupContentClonerManager $cloner_manager,
    EntityTypeManagerInterface $entity_type_manager,
    StateInterface $state
  ) {
    $this->clonerManager = $cloner_manager;
    $this->clonerPlugins = ['_default' => $cloner_manager->createInstance('default')];
    $this->entityTypeManager = $entity_type_manager;
    $this->state = $state;
  }

  /**
   * Allow other modules to check if cloning is in progress.
   */
  public function isCloning(): bool {
    return $this->cloning;
  }

  /**
   * Allow other modules to check if reverting is in progress.
   */
  public function isReverting(): bool {
    return $this->reverting;
  }

  /**
   * Sets cloning settings by content type.
   *
   * @param int[] $settings
   *   Settings.
   * @param Drupal\group\Entity\GroupInterface $group
   *   Group.
   */
  public function setCloningSettings(array $settings = NULL, GroupInterface $group = NULL): void {
    if ($settings === NULL && !$group instanceof GroupInterface) {
      throw new \InvalidArgumentException('Settings or group needs to be provided in setCloningSettings method.');
    }

    if ($settings === NULL) {
      $config_entities = $this->entityTypeManager->getStorage('group_clone_settings')->loadByProperties([
        'status' => TRUE,
        'id' => $group->bundle(),
      ]);
      if (!count($config_entities)) {
        throw new \InvalidArgumentException(sprintf('No settings found for %s group type.', $group->bundle()));
      }
      $this->settings = $config_entities[$group->bundle()]->get('content_types');
    }
    else {
      $this->settings = $settings;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setClonedEntitityIds(array $cloned_entity_ids): void {
    $this->clonedEntityIDs = $cloned_entity_ids;
  }

  /**
   * {@inheritdoc}
   */
  public function getClonedEntitityIds(): array {
    return $this->clonedEntityIDs;
  }

  /**
   * Starts the group cloning operation.
   *
   * @return \Drupal\group\Entity\GroupInterface
   *   Cloned group or NULL if a batch operation has been initialized or
   *   errors occurred.
   */
  public function clone(GroupInterface $group, bool $use_batch_api = TRUE): ?GroupInterface {
    $content_ids = [];
    if (\count($this->settings)) {
      $content_types = [];
      foreach ($this->settings as $type_id => $value) {
        if ($value !== 'disabled') {
          $content_types[] = $type_id;
        }
      }
      if (\count($content_types)) {
        $query = $this->entityTypeManager->getStorage('group_content')->getQuery();
        $query->accessCheck(FALSE);
        $query->condition('gid', $group->id());
        $query->condition('type', $content_types, 'IN');
        $content_ids = $query->execute();
      }
    }

    // Clone the group with all referenced entities first.
    // Override label and status of the group.
    $this->cloning = TRUE;
    $group_clone = $this->cloneEntity($group, $this->clonedEntityIDs, 'all', [
      'label' => $group->label() . ' - Cloned',
      'status' => FALSE,
    ]);
    $this->cloning = FALSE;

    if (!count($content_ids)) {
      return $group_clone;
    }

    if ($use_batch_api) {
      $data = [
        'original_group_id' => $group->id(),
        'group_clone_id' => $group_clone->id(),
        'group_label' => $group->label(),
        'content_ids' => $content_ids,
        'settings' => $this->settings,
        'cloned_entity_ids' => $this->clonedEntityIDs,
      ];
      $batch = GroupClonerBatch::getCloneBatch($data);
      \batch_set($batch);
      return NULL;
    }
    else {
      try {
        foreach ($content_ids as $content_id) {
          $this->cloneContent($content_id, $group_clone->id());
        }
        $this->storeClonedData($group->id(), $group_clone->id());
        return $group_clone;
      }
      catch (\Exception $e) {
        $this->storeClonedData($group->id(), $group_clone->id());
        throw $e;
      }
    }

    return NULL;
  }

  /**
   * Revert single item.
   */
  public function revertSingle(string $entity_type_id, string $entity_id): void {
    $this->reverting = TRUE;
    $entity = $this->entityTypeManager->getStorage($entity_type_id)->load($entity_id);
    if ($entity instanceof EntityInterface) {
      $entity->delete();
    }
    $this->reverting = FALSE;
  }

  /**
   * Revert cloning.
   */
  public function revert(string $cloned_group_id, bool $use_batch_api): ?Url {
    $group_storage = $this->entityTypeManager->getStorage('group');
    $cloned_group = $group_storage->load($cloned_group_id);
    $data = $this->state->get(self::STATE_PREFIX . $cloned_group_id);
    if (!$cloned_group instanceof GroupInterface || $data === NULL) {
      throw new \InvalidArgumentException('Invalid cloned group ID provided in GroupCloner::revert()');
    }
    $redirect_url = $group_storage->load($data['gid'])->toUrl();
    if ($use_batch_api) {
      $data['group_label'] = $cloned_group->label();
      $data['redirect_url'] = $redirect_url->toString();
      $data['group_clone_id'] = $cloned_group->id();
      $batch = GroupClonerBatch::getRevertBatch($data);
      \batch_set($batch);
    }
    else {
      foreach ($data['cloned_entity_ids'] as $entity_type_id => $type_entity_ids) {
        foreach ($type_entity_ids as $entity_id) {
          $this->revertSingle($entity_type_id, $entity_id);
        }
      }
      return $redirect_url;
    }
    return NULL;
  }

  /**
   * Gets cloner plugin from entity.
   */
  private function getClonerPlugin(GroupContentInterface $group_content): GroupContentClonerInterface {
    if (!array_key_exists($group_content->bundle(), $this->pluginMap)) {
      $this->pluginMap[$group_content->bundle()] = $group_content->getContentPlugin()->getBaseId();
    }

    $plugin_base_id = $this->pluginMap[$group_content->bundle()];
    if (!\array_key_exists($plugin_base_id, $this->clonerPlugins)) {
      $definitions = $this->clonerManager->getDefinitionsByPlugin($plugin_base_id);
      // We don't expect to have more than 1 here.
      if (\count($definitions) > 1) {
        throw new \Exception(sprintf('Multiple cloner plugins found (%s) for the %s group content plugin', \implode(', '), \array_keys($definitions)), $plugin_base_id);
      }
      if (\count($definitions) === 1) {
        $definition = \reset($definitions);
        $this->clonerPlugins[$plugin_base_id] = $this->clonerManager->createInstance($definition['id']);
      }
      else {
        $this->clonerPlugins[$plugin_base_id] = NULL;
      }
    }

    // We have all the data built, just return dedicated or default.
    if ($this->clonerPlugins[$plugin_base_id] instanceof GroupContentClonerInterface) {
      return $this->clonerPlugins[$plugin_base_id];
    }
    return $this->clonerPlugins['_default'];
  }

  /**
   * Clone group content.
   */
  public function cloneContent(string $entity_id, string $group_clone_id): void {
    $group_content = $this->entityTypeManager->getStorage('group_content')->load($entity_id);
    if (!$group_content instanceof GroupContentInterface) {
      return;
    }

    $plugin = $this->getClonerPlugin($group_content);
    $this->cloning = TRUE;
    $plugin->clone($group_content, $group_clone_id, $this->settings[$group_content->bundle()], $this->clonedEntityIDs);
    $this->cloning = FALSE;
  }

  /**
   * Store cloned data in state to allow reverting when necessary.
   */
  public function storeClonedData(string $original_group_id, string $cloned_group_id, ?array $cloned_entity_ids = NULL) {
    if ($cloned_entity_ids === NULL) {
      $cloned_entity_ids = $this->clonedEntityIDs;
    }
    $this->state->set(self::STATE_PREFIX . $cloned_group_id, [
      'gid' => $original_group_id,
      'cloned_entity_ids' => $cloned_entity_ids,
    ]);
  }

  /**
   * Helper function to get cloning results.
   *
   * @param mixed[] $results
   *   Array of results.
   * @param throwable $exception
   *   The execption thrown during cloning or NULL.
   */
  public function getResultsMessage(array $results, throwable $exception = NULL): TranslatableMarkup {
    if ($exception instanceof \Throwable) {
      return $this->t('Finished with an error (@message), please inspect the result group and revert if desired.', [
        '@message' => $exception->getMessage(),
      ]);
    }

    if (!array_key_exists('group_label', $results)) {
      throw new \InvalidArgumentException('Missing group_label array key in getResultsMessage argument.');
    }
    if (!array_key_exists('group_clone_id', $results)) {
      throw new \InvalidArgumentException('Missing group_clone_id array key in getResultsMessage argument.');
    }
    if (!array_key_exists('cloned_entity_ids', $results)) {
      $results['cloned_entity_ids'] = $this->clonedEntityIDs;
    }

    $details = [];
    foreach ($results['cloned_entity_ids'] as $entity_type_id => $entity_ids) {
      $details[] = $entity_type_id . ': ' . \count($entity_ids);
    }

    $cloned_group = $this->entityTypeManager->getStorage('group')->load($results['group_clone_id']);
    return $this->t('Cloned group "@label" to "@cloned_label". Entities cloned: @details', [
      '@label' => $results['group_label'],
      '@cloned_label' => $cloned_group->label(),
      '@details' => \count($details) ? \implode(', ', $details) : '0',
    ]);
  }

}

==> group_clone-3.0.x-dev.tar.gz: group_clone/src/GroupCloner.php <==
<?php

declare(strict_types=1);

namespace Drupal\group_clone;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\SynchronizableInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\group\Entity\GroupInterface;
use Drupal\group_clone\Plugin\GroupContentCloner\GroupContentClonerInterface;

/**
 * {@inheritdoc}
 */
final class GroupCloner {

  use EntityCloneTrait;
  use StringTranslationTrait;

  public const STATE_PREFIX = 'group_clone_data.';
  public const PRESERVE_VALUE = '#preserve#';

  /**
   * Cloning settings keyed by group content type.
   *
   * @var int[]
   */
  private array $settings = [];

  /**
   * Group content type - base plugin ID map.
   *
   * Data caching purpose.
   *
   * @var string[]
   */
  private array $pluginMap = [];

  /**
   * Cloner plugins.
   *
   * @var \Drupal\group_clone\Plugin\GroupContentCloner\GroupContentClonerInterface[]
   */
  private array $clonerPlugins = [];

  /**
   * Cloned entities array for referencing.
   *
   * Keyed by entity type ID as it can contain all types of entities
   * and not only group_content ones.
   *
   * @var mixed[]
   */
  private array $clonedEntityIDs = [];

  /**
   * Constructs a new GroupCloner.
   */
  public function __construct(
    private readonly GroupContentClonerManager $clonerManager,
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly AccountProxyInterface $currentUser,
    private readonly StateInterface $state
  ) {}

  /**
   * Sets cloning settings by content type.
   *
   * @param mixed[] $settings
   *   Settings by plugin.
   */
  public function setCloningSettings(array $settings): void {
    if (!\array_key_exists('field_config', $settings)) {
      $settings['field_config'] = [];
    }
    $this->settings = $settings;
  }

  /**
   * {@inheritdoc}
   */
  public function setClonedEntitityIds(array $cloned_entity_ids): void {
    $this->clonedEntityIDs = $cloned_entity_ids;
  }

  /**
   * {@inheritdoc}
   */
  public function getClonedEntitityIds(): array {
    return $this->clonedEntityIDs;
  }

  /**
   * Starts the group cloning operation.
   *
   * @return \Drupal\group\Entity\GroupInterface
   *   Cloned group or NULL if a batch operation has been initialized or
   *   errors occurred.
   */
  public function clone(GroupInterface $group, bool $use_batch_api = TRUE): ?GroupInterface {

    $content_items = [];
    foreach ($this->settings['plugin_config'] as $plugin_id => $plugin_config) {
      $plugin = $this->getClonerPlugin($plugin_id, $plugin_config);
      $plugin_content_ids = $plugin->getContentIdsToClone($group);
      foreach ($plugin_content_ids as $type => $ids) {
        foreach ($ids as $id) {
          $content_items[] = [$plugin_id, $type, $id];
        }
      }
    }

    $field_values = [];

    // Clone the group respecting referenced entities settings first.
    foreach ($this->settings['field_config'] as $field_id => $value) {
      if (!$value) {
        $field_values[$field_id] = self::PRESERVE_VALUE;
      }
    }

    // Override label and status of the group.
    $field_values['label'] = $group->label() . ' - Cloned';
    $field_values['status'] = FALSE;
    // On group save, owner membership may be added by default. Currently
    // it's not possible to set isSyncing = TRUE on that entity. Set owner
    // to the current user to minimize possible impact.
    $field_values['uid'] = $this->currentUser->id();

    $group_clone = $this->cloneEntity($group, $this->clonedEntityIDs, 'all', $field_values);

    if (!count($content_items)) {
      return $group_clone;
    }

    if ($use_batch_api) {
      $data = [
        'original_group_id' => $group->id(),
        'group_clone_id' => $group_clone->id(),
        'group_label' => $group->label(),
        'content_items' => $content_items,
        'settings' => $this->settings,
        'cloned_entity_ids' => $this->clonedEntityIDs,
      ];
      $batch = GroupClonerBatch::getCloneBatch($data);
      \batch_set($batch);
      return NULL;
    }
    else {
      try {
        foreach ($content_items as $item) {
          $this->cloneItem($item, $group_clone->id());
        }
        $this->storeClonedData($group->id(), $group_clone->id());
        return $group_clone;
      }
      catch (\Exception $e) {
        $this->storeClonedData($group->id(), $group_clone->id());
        throw $e;
      }
    }

    return NULL;
  }

  /**
   * Revert single item.
   */
  public function revertSingle(string $entity_type_id, string $entity_id): void {
    $entity = $this->entityTypeManager->getStorage($entity_type_id)->load($entity_id);
    if ($entity instanceof EntityInterface) {
      if ($entity instanceof SynchronizableInterface) {
        $entity->setSyncing(TRUE);
      }
      $entity->delete();
    }
  }

  /**
   * Revert cloning.
   */
  public function revert(string $cloned_group_id, bool $use_batch_api): ?Url {
    $group_storage = $this->entityTypeManager->getStorage('group');
    $cloned_group = $group_storage->load($cloned_group_id);
    $data = $this->state->get(self::STATE_PREFIX . $cloned_group_id);
    if (!$cloned_group instanceof GroupInterface || $data === NULL) {
      throw new \InvalidArgumentException('Invalid cloned group ID provided in GroupCloner::revert()');
    }
    $redirect_url = $group_storage->load($data['gid'])->toUrl();
    if ($use_batch_api) {
      $data['group_label'] = $cloned_group->label();
      $data['redirect_url'] = $redirect_url->toString();
      $data['group_clone_id'] = $cloned_group->id();
      $batch = GroupClonerBatch::getRevertBatch($data);
      \batch_set($batch);
    }
    else {
      foreach ($data['cloned_entity_ids'] as $entity_type_id => $type_entity_ids) {
        foreach ($type_entity_ids as $entity_id) {
          $this->revertSingle($entity_type_id, $entity_id);
        }
      }
      return $redirect_url;
    }
    return NULL;
  }

  /**
   * Gets cloner plugin from entity.
   */
  private function getClonerPlugin(string $plugin_id, array $plugin_config): GroupContentClonerInterface {
    if (!\array_key_exists($plugin_id, $this->clonerPlugins)) {
      $this->clonerPlugins[$plugin_id] = $this->clonerManager->createInstance($plugin_id, $plugin_config);
    }
    return $this->clonerPlugins[$plugin_id];
  }

  /**
   * Clone group content.
   */
  public function cloneItem(array $item, string $group_clone_id): void {
    [$plugin_id, $type, $entity_id] = $item;
    $plugin = $this->getClonerPlugin($plugin_id, $this->settings['plugin_config'][$plugin_id]);
    $plugin->clone($type, $entity_id, $group_clone_id, $this->clonedEntityIDs);
  }

  /**
   * Store cloned data in state to allow reverting when necessary.
   */
  public function storeClonedData(string $original_group_id, string $cloned_group_id, ?array $cloned_entity_ids = NULL) {
    if ($cloned_entity_ids === NULL) {
      $cloned_entity_ids = $this->clonedEntityIDs;
    }
    $this->state->set(self::STATE_PREFIX . $cloned_group_id, [
      'gid' => $original_group_id,
      'cloned_entity_ids' => $cloned_entity_ids,
    ]);
  }

  /**
   * Helper function to get cloning results.
   *
   * @param mixed[] $results
   *   Array of results.
   * @param throwable $exception
   *   The execption thrown during cloning or NULL.
   */
  public function getResultsMessage(array $results, \Throwable $exception = NULL): TranslatableMarkup {
    if ($exception instanceof \Throwable) {
      return $this->t('Finished with an error (@message), please inspect the result group and revert if desired.', [
        '@message' => $exception->getMessage(),
      ]);
    }

    if (!array_key_exists('group_label', $results)) {
      throw new \InvalidArgumentException('Missing group_label array key in getResultsMessage argument.');
    }
    if (!array_key_exists('group_clone_id', $results)) {
      throw new \InvalidArgumentException('Missing group_clone_id array key in getResultsMessage argument.');
    }
    if (!array_key_exists('cloned_entity_ids', $results)) {
      $results['cloned_entity_ids'] = $this->clonedEntityIDs;
    }

    $details = [];
    foreach ($results['cloned_entity_ids'] as $entity_type_id => $entity_ids) {
      $details[] = $entity_type_id . ': ' . \count($entity_ids);
    }

    $cloned_group = $this->entityTypeManager->getStorage('group')->load($results['group_clone_id']);
    return $this->t('Cloned group "@label" to "@cloned_label". Entities cloned: @details', [
      '@label' => $results['group_label'],
      '@cloned_label' => $cloned_group->label(),
      '@details' => \count($details) ? \implode(', ', $details) : '0',
    ]);
  }

}
